==> tecfy_core_package-1.0.15.tar.gz: lib/controllers/entity/entity_form.controller.dart <==
part of tecfy_core_client_package;

class EntityFormController extends BaseController {
  late Entity entity;
  late String titleField;
  late String subTitleField;
  late String tailField;

  /// data
  late Map<String, dynamic>? data;

  /// properties
  late List<Property> properties;

  /// properties need calculation
  late List<Property> calcProps;

  /// is new object
  bool isAddNew = false;

  /// is true, the controller will load data from server on open
  bool loadData = false;

  /// form tabs
  late Map<String, List<List<Property>>> tabs;
  bool get canSave =>
      !_readOnly &&
      (data != null && data!['isLocked'] != true) &&
      ((isAddNew && entity.permission.canAdd) ||
          (!isAddNew && entity.permission.canUpdate));
  bool _readOnly;
  bool get readOnly => !canSave;
  List<dynamic>? _entitySaveConfirm;

  EntityFormController(State<StatefulWidget> state, String entityName,
      this.data, this._readOnly, this.loadData)
      : super(state) {
    busy = false;

    print('-------------- EntityFormController => $entityName');
    entity = EntityService.getEntity(entityName) ?? Entity.fromMap({});
    print(
        'Entity ${entity.dbName} Can Add: ${entity.permission.canAdd} Can Edit: ${entity.permission.canUpdate}');
    properties =
        entity.properties.where((p) => isAddNew ? p.isAdd : p.isEdit).toList();
    //properties.forEach((p) => p.isVisable = true);
    calcProps = properties
        .where((p) =>
            p.propertyType == PropertyTypes.Number &&
            // (p as PropertyNumber).calcFormula != null &&
            (p as PropertyNumber).calcFormula.isNotEmpty)
        .toList();
    isAddNew = data == null || data!['id'] == null;
    if (data == null) {
      data = Map<String, dynamic>();
    }
    initTabs();
    loadFromDb();
  }

  loadFromDb() async {
    if (entity.isLocal || !loadData) return;
    if (data == null || data!['id'] == null) {
      await onValueChanged('', '', true); // for default values
      return;
    }
    await Future.delayed(Duration(
        milliseconds:
            30)); // delay a bit because you cannot show "loading" before page initlization
    setLoading();
    var result = await EntityService.getOne(entity.dbName, data!['id'],
        forView: readOnly, withRelatedData: true);
    setResume();
    if (result != null) {
      // print('------------- got(loadFromDb) -----------');
      data = result;
      print(data);
      await onValueChanged('', '', true);
      setState();
    }
  }

  onValueChanged(value, propName, [calcAll = false]) async {
    if (readOnly) return;
    print('-----------value changed ! ($propName) $calcAll');
    try {
      if (!calcAll) data![propName] = value;

      bool updated = await entity.setDefaultValues(data!);

      // for (Property p in properties) {
      //   var r = await p.isVisible(data, entity);
      //   if (r != p.isVisable) {
      //     p.isVisable = r;
      updated = true;
      //   }
      // }
      // check the calced properties
      for (PropertyNumber p in calcProps
          .where((e) => e.propertyType == PropertyTypes.Number)
          .toList()
          .cast<PropertyNumber>()) {
        if (p.formulaType == "Formula") {
          if (calcAll ||
              p.calcedProps.contains(p.upperFirst(propName)) ||
              p.calcedProps.contains(propName)) {
            var vars = p.calcedProps
                .map((pro) => data![p.lowerFirst(pro)] ?? 0)
                .toList();
            await onValueChanged(p.getCalcResult(vars), p.dbName);
            setState();
            print('calc ${p.dbName}');
          }
        } else if (p.formulaType == "SumSub") {
          if (calcAll || p.sumEntityName == propName) {
            double val = 0;
            if (data![p.sumEntityName] != null &&
                data![p.sumEntityName] is List) {
              for (var row in data![p.sumEntityName]) {
                val += row[p.sumPropertyName] ?? 0;
              }
            }
            await onValueChanged(val, p.dbName);
            setState();
          }
        }
      }

      if (updated // || (!calcAll && properties.any((p) => p.filterValueField == propName))
          ) setState();
    } catch (err) {
      print('XXXXXX ERROR onValueChanged: $err');
    }
  }

  Future<bool> onSave() async {
    if (readOnly) return false;
    // show loading
    setLoading();
    var result = await EntityService.saveEntity(entity, data, context, isAddNew,
        entitySaveConfirm: _entitySaveConfirm);
    setResume();
    if (!result.success) {
      try {
        print(result.data);
        var question = '';
        if (result.data != null && result.data.length > 0) {
          question = result.data[0]['nameAr'] ?? '';
        }
        if (question.isNotEmpty) {
          var r = await AppNavigator.confirmMessage(
              context, result.message ?? '',
              message: question);
          if (!r) return false;
          result.data[0]['value'] = true;
          _entitySaveConfirm = result.data;
          return await onSave();
        } else {
          AppNavigator.showMessage(context, result.message ?? '',
              type: MessageType.error, message: question);
        }
      } catch (err, t) {
        print(err);
        print(t);
      }
      return false;
    }
    ConfigService.setValueMap("savedDataForm${entity.name}", {});
    // show message
    await AppNavigator.showMessage(context, 'تم الحفظ بنجاح',
        type: MessageType.info);
    AppNavigator.pop(context, true);
    return true;
  }

  Future<bool> onSaveNew() async {
    if (readOnly) return false;
    // show loading
    isAddNew = true;
    resetIds(data!);
    return await onSave();
  }

  resetIds(Map<String, dynamic> dat) {
    dat['id'] = null;
    dat['recordNumber'] = null;
    for (var k in dat.keys) {
      if (dat[k] is List) {
        print('d---------$k');
        for (var d in dat[k]) {
          resetIds(d);
        }
      }
    }
  }

  lowerFirst(String s) {
    return '${s[0].toLowerCase()}${s.substring(1)}';
  }

  void initTabs() {
    this.tabs = {};
    for (var prop in this.entity.properties) {
      if (isAddNew ? prop.isAdd : prop.isEdit) {
        if (prop.editTabName == null || (prop.editTabName?.isEmpty ?? true))
          prop.editTabName = AppLocale.translate(context, 'Details');

//prop.propertyType == PropertyTypes.OtherEntityMany
        if (prop.size == PropertySizes.NewTab) {
          this.tabs[prop.name] = [
            [prop]
          ];
        } else {
          if (this.tabs.length == 0)
            this.tabs[prop.editTabName ?? ''] = [
              [prop]
            ];
          else {
            if (this.tabs[prop.editTabName] == null) {
              this.tabs[prop.editTabName ?? ''] = [
                [prop]
              ];
            } else {
              if (prop.fullWidth)
                this.tabs[prop.editTabName]?.add([prop]);
              else
                this.tabs[prop.editTabName]?.last.add(prop);
            }
          }
        }
      }
      if (prop.newLineAfter) this.tabs[prop.editTabName]?.add([]);
    }

    for (var tk in tabs.keys) {
      tabs[tk] = tabs[tk]?.where((e) => e.length > 0).toList() ?? [];
      if (tabs[tk]?.length == 0) {
        tabs.remove(tk);
      }
    }
  }
}

==> tecfy_core_package-1.0.16.tar.gz: lib/controllers/entity/entity_form.controller.dart <==
part of tecfy_core_client_package;

class EntityFormController extends BaseController {
  late Entity entity;
  late String titleField;
  late String subTitleField;
  late String tailField;

  /// data
  late Map<String, dynamic>? data;

  /// properties
  late List<Property> properties;

  /// properties need calculation
  late List<Property> calcProps;

  /// is new object
  bool isAddNew = false;

  /// is true, the controller will load data from server on open
  bool loadData = false;

  /// form tabs
  late Map<String, List<List<Property>>> tabs;
  bool get canSave =>
      !_readOnly &&
      (data != null && data!['isLocked'] != true) &&
      ((isAddNew && entity.permission.canAdd) ||
          (!isAddNew && entity.permission.canUpdate));
  bool _readOnly;
  bool get readOnly => !canSave;
  List<dynamic>? _entitySaveConfirm;

  EntityFormController(State<StatefulWidget> state, String entityName,
      this.data, this._readOnly, this.loadData)
      : super(state) {
    busy = false;

    print('-------------- EntityFormController => $entityName');
    entity = EntityService.getEntity(entityName) ?? Entity.fromMap({});
    print(
        'Entity ${entity.dbName} Can Add: ${entity.permission.canAdd} Can Edit: ${entity.permission.canUpdate}');
    properties =
        entity.properties.where((p) => isAddNew ? p.isAdd : p.isEdit).toList();
    //properties.forEach((p) => p.isVisable = true);
    calcProps = properties
        .where((p) =>
            p.propertyType == PropertyTypes.Number &&
            // (p as PropertyNumber).calcFormula != null &&
            (p as PropertyNumber).calcFormula.isNotEmpty)
        .toList();
    isAddNew = data == null || data!['id'] == null;
    if (data == null) {
      data = Map<String, dynamic>();
    }
    initTabs();
    loadFromDb();
  }

  loadFromDb() async {
    if (entity.isLocal || !loadData) return;
    if (data == null || data!['id'] == null) {
      await onValueChanged('', '', true); // for default values
      return;
    }
    await Future.delayed(Duration(
        milliseconds:
            30)); // delay a bit because you cannot show "loading" before page initlization
    setLoading();
    var result = await EntityService.getOne(entity.dbName, data!['id'],
        forView: readOnly, withRelatedData: true);
    setResume();
    if (result != null) {
      // print('------------- got(loadFromDb) -----------');
      data = result;
      print(data);
      await onValueChanged('', '', true);
      setState();
    }
  }

  onValueChanged(value, propName, [calcAll = false]) async {
    if (readOnly) return;
    print('-----------value changed ! ($propName) $calcAll');
    try {
      if (!calcAll) data![propName] = value;

      bool updated = await entity.setDefaultValues(data!);

      // for (Property p in properties) {
      //   var r = await p.isVisible(data, entity);
      //   if (r != p.isVisable) {
      //     p.isVisable = r;
      updated = true;
      //   }
      // }
      // check the calced properties
      for (PropertyNumber p in calcProps
          .where((e) => e.propertyType == PropertyTypes.Number)
          .toList()
          .cast<PropertyNumber>()) {
        if (p.formulaType == "Formula") {
          if (calcAll ||
              p.calcedProps.contains(p.upperFirst(propName)) ||
              p.calcedProps.contains(propName)) {
            var vars = p.calcedProps
                .map((pro) => data![p.lowerFirst(pro)] ?? 0)
                .toList();
            await onValueChanged(p.getCalcResult(vars), p.dbName);
            setState();
            print('calc ${p.dbName}');
          }
        } else if (p.formulaType == "SumSub") {
          if (calcAll || p.sumEntityName == propName) {
            double val = 0;
            if (data![p.sumEntityName] != null &&
                data![p.sumEntityName] is List) {
              for (var row in data![p.sumEntityName]) {
                val += row[p.sumPropertyName] ?? 0;
              }
            }
            await onValueChanged(val, p.dbName);
            setState();
          }
        }
      }

      if (updated // || (!calcAll && properties.any((p) => p.filterValueField == propName))
          ) setState();
    } catch (err, t) {
      print('XXXXXX ERROR onValueChanged: $err : t');
    }
  }

  Future<bool> onSave() async {
    if (readOnly) return false;
    // show loading
    setLoading();
    var result = await EntityService.saveEntity(entity, data, context, isAddNew,
        entitySaveConfirm: _entitySaveConfirm);
    setResume();
    if (!result.success) {
      try {
        print(result.data);
        var question = '';
        if (result.data != null && result.data.length > 0) {
          question = result.data[0]['nameAr'] ?? '';
        }
        if (question.isNotEmpty) {
          var r = await AppNavigator.confirmMessage(
              context, result.message ?? '',
              message: question);
          if (!r) return false;
          result.data[0]['value'] = true;
          _entitySaveConfirm = result.data;
          return await onSave();
        } else {
          AppNavigator.showMessage(context, result.message ?? '',
              type: MessageType.error, message: question);
        }
      } catch (err, t) {
        print(err);
        print(t);
      }
      return false;
    }
    ConfigService.setValueMap("savedDataForm${entity.name}", {});
    // show message
    await AppNavigator.showMessage(context, 'تم الحفظ بنجاح',
        type: MessageType.info);
    AppNavigator.pop(context, true);
    return true;
  }

  Future<bool> onSaveNew() async {
    if (readOnly) return false;
    // show loading
    isAddNew = true;
    resetIds(data!);
    return await onSave();
  }

  resetIds(Map<String, dynamic> dat) {
    dat['id'] = null;
    dat['recordNumber'] = null;
    for (var k in dat.keys) {
      if (dat[k] is List) {
        print('d---------$k');
        for (var d in dat[k]) {
          resetIds(d);
        }
      }
    }
  }

  lowerFirst(String s) {
    return '${s[0].toLowerCase()}${s.substring(1)}';
  }

  void initTabs() {
    this.tabs = {};
    for (var prop in this.entity.properties) {
      if (isAddNew ? prop.isAdd : prop.isEdit) {
        if (prop.editTabName == null || (prop.editTabName?.isEmpty ?? true))
          prop.editTabName = AppLocale.translate(context, 'Details');

//prop.propertyType == PropertyTypes.OtherEntityMany
        if (prop.size == PropertySizes.NewTab) {
          this.tabs[prop.name] = [
            [prop]
          ];
        } else {
          if (this.tabs.length == 0)
            this.tabs[prop.editTabName ?? ''] = [
              [prop]
            ];
          else {
            if (this.tabs[prop.editTabName] == null) {
              this.tabs[prop.editTabName ?? ''] = [
                [prop]
              ];
            } else {
              if (prop.fullWidth)
                this.tabs[prop.editTabName]?.add([prop]);
              else
                this.tabs[prop.editTabName]?.last.add(prop);
            }
          }
        }
      }
      if (prop.newLineAfter) this.tabs[prop.editTabName]?.add([]);
    }

    for (var tk in tabs.keys) {
      tabs[tk] = tabs[tk]?.where((e) => e.length > 0).toList() ?? [];
      if (tabs[tk]?.length == 0) {
        tabs.remove(tk);
      }
    }
  }
}
